#include <bit>
#include <system_error>

#include <errno.h>

#include <mutex.hh>
#include <condvar.hh>
#include <queue.hh>

//#include <signal.hh> // NOTE: Moved to `fx-sig.cpp`

#include "macros.h"

namespace fx::error {
	OperationCancelledError::OperationCancelledError() noexcept
			: std::runtime_error("Operation cancelled") {}
	TaskCancelledError::TaskCancelledError() noexcept
			: std::runtime_error("Operation cancelled")
			, OperationCancelledError() {}

	OperationCancelledError::~OperationCancelledError() = default;
	TaskCancelledError::~TaskCancelledError() = default;

	[[gnu::noinline, gnu::cold, noreturn]]
	void throw_cancelled(bool clean)
	{
		if(clean) throw TaskCancelledError();
		else	throw OperationCancelledError();
	}

	void throw_if_cancelled(std::stop_token const& st, bool clean) {
		if(__builtin_expect(st.stop_requested(), false))
			throw_cancelled(clean);
	}

	[[gnu::cold, noreturn]]
	void throw_last_system_error() {
		throw std::system_error{ int(errno), std::generic_category() };
	}
}
