#ifndef _OTPS_IMPL_MACROS_H
#define _OTPS_IMPL_MACROS_H

#ifdef __cplusplus /* C++ only */
extern "C" {

#define namespace$static namespace 
#define namespace$static$IN(N) namespace N { namespace

#define BLOCK$static(N) namespace$static$AT(N) {
#define END$static } }

#define namespace$static$inline(V) namespace$static [[gnu::__visibility__( #V )]]
#define BLOCK$static$inline(V) namespace$static$inline(V) {
#define END$static$inline }

#define namespace$extern extern "C"
#define BLOCK$extern namespace$extern {
#define END$extern }

#define throw$error(X, ...) [] [[noreturn, gnu::noinline, gnu::cold]] (auto&&... ctor) static { throw ::std::decay_t<__typeof__(X)> { ::std::forward<decltype(ctor)>(ctor)... }; } \
	(__VA_ARGS__)

#else /* C only */

//NOTE: No replacement for `namespace$static[$inline]`: It cannot be affected in C syntax (without `#pragma GCC`.)

#define namespace$extern /* default behaviour in C */
#define BLOCK$extern
#define END$extern

#endif /* C & C++ */

#define throw$extern(_message) throw_last_system_error()
extern void throw_last_system_error() __attribute__((__visibility__("internal"), __noinline__, __cold__, noreturn));

#define ASSUME(X) do { if(!(X)) __builtin_unreachable(); } while(0)

#define EXPECT(X, V) __builtin_expect(!!(X), !!(V))

#define LIKELY(X) EXPECT((X), 1)
#define UNLIKELY(X) EXPECT((X), 0)

#define ASSUME_NONNULL(p) ((p) ?: __builtin_unreachable())
#define ASSERT_NONNULL(p) ((p) ?: __builtin_trap())

//#define IMPL_CLASS N
#define IMPL_SPEC $NONE /*<T>*/

#define $class IMPL_CLASS

#define $ $class IMPL_SPEC ::

#define $ctor $ $class
#define $dtor $ ~ $class

#define $ctor_move(N) $ctor($class && N)
#define $ctor_copy(N) $ctor($class const& N)

#define $assign $class & $ operator=

#define $assign_move(N) $assign($class && N)
#define $assign_copy(N) $assign($class const& N)

#define this$not(N) (__builtin_expect(this != ::std::addressof(N), true))

/// OOL-defaults a class's destructor. (Useful for un-`inline`ing non-`constexpr` virtual dtors.
///
/// Doing this keeps the vtable definition within this TU, and reduces linker's redundency strain of generating multiple vtables.
///
/// Also since this TU has access to all non-`inline`, non-`constexpr` types and functions used within that class, the compiler can optimise more around the vtable and what happens in the destructor (de-virtualisation and pre-LTO inlining.)
#define class$default(X, ...) X::~X() __VA_ARGS__ = default

//#undef IMPL_CLASS

#ifdef __cplusplus
}
#endif

#endif /* _OTPS_IMPL_MACROS_H */
